﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Saleae.SocketApi;
using System.IO;
using System.Globalization;
using System.IO.Ports;
using NDesk.Options;
using System.Text.RegularExpressions;

namespace SaleaeTools
{
    class Program
    {

        static void OptionParseBool(string value, ref bool target)
        {
            value = value.ToUpper();

            // parse values like y, yes, t, true, n, no, f, false, etc.
            // if the value is set but empty, handle as true - because the parameter is set

            if (value.StartsWith("Y") || value.StartsWith("T"))
                target = true;
            else if (value.StartsWith("N") || value.StartsWith("F"))
                target = false;
            else if (value != null)
                target = true;
        }

        static void Main(string[] args)
        {
            bool show_help = false;
            string saleae_export = null;
            string saleae_dir = null;
            bool saleae_exportlist = false;
            string saleae_ip = "127.0.0.1";
            int saleae_port = 10429;
            
            string protocol = "";
            string ifname = "";
            string ofname = "";
            string out_colsep = ";";
            bool out_colsep_hint = false;
            string out_datasep = "";
            bool out_localdecimal = false;
            bool out_striphexprefix = false;
            bool out_convertAscii = false;
            double septime = 0;
            string combine = null;

            OptionSet options = new OptionSet()
            {
                { "h|help", "Show this help and exit", v => OptionParseBool(v, ref show_help) },

                { "el|exportlist", "Lists the exportable analyzers", v => OptionParseBool(v, ref saleae_exportlist) },
                { "e|export=", "Exports selected (comma separated) or all (*) analyzers from Logic. Scripting socket server must be enabled.", v => saleae_export = v },
                
                { "ed|exportdir", "Lists the exportable analyzers", v => saleae_dir = v },
                { "si|saleaeip=", "IP of the logic client (default: 127.0.0.1)", v => saleae_ip = v },
                { "sp|saleaeport=", "Port of the logic client (default: 10429)", (int v) => saleae_port = v },

                { "c|combine=", "Combines multiple analyzer export files. Also works with hex compressed files w/o local decimal", v => combine = v },

                { "p|protocol=", "Protocol of the analyzer export file (SERial|I2C|SPI)", v => protocol = v },
                { "if|infile=", "Input file", v => ifname = v },
                { "of|outfile=", "Output file (optional)", v => ofname = v },
                { "cs|colsep=", "Column separator for output file (default: ;)", v => out_colsep = v },
                { "csh|colsephint", "Hint column separator in CSV-file",v => OptionParseBool(v, ref out_colsep_hint) },
                { "ds|datasep=", "Data separator (default: nothing)", v => out_datasep = v },
                { "ld|localdecimal", "Use local decimal separator for timestamps (default: off)", v => OptionParseBool(v, ref out_localdecimal) },
                { "sh|striphex", "Strips hexadecimal prefix \"0x\"", v => out_striphexprefix = v != null },
                { "ca|convertascii", "Convert Saleae's way of handling ASCII (control characters, COMMA)", v => OptionParseBool(v, ref out_convertAscii) },
                { "st|septime=", "Separation time in ms where separate packets are detected (Serial only)", v => septime = double.Parse(v, CultureInfo.InvariantCulture) }
            };

            List<string> extra = null;
            try
            {
                extra = options.Parse(args);
            }
            catch (OptionException ex)
            {
                Console.WriteLine("Error: {0}", ex.Message);
                Console.WriteLine("Try 'serialplayer --help' for more information");
                return;
            }

            if (show_help || args.Length == 0)
            {
                string exename = AppDomain.CurrentDomain.FriendlyName;
                Console.WriteLine("Tool to process Saleae Logic analyzer export files");
                Console.WriteLine();
                options.WriteOptionDescriptions(Console.Out);

                Console.WriteLine("Export analyzer results from Saleae Logic");
                Console.WriteLine("{0} -e (1,2,3|*) [-ed dir] [-si ip] [-sp port]", exename);
                Console.WriteLine();
                Console.WriteLine("Combine multiple files:");
                Console.WriteLine("{0} -c outfile.csv [-csh] infile1.txt infile2.txt [infile3.txt ...]", exename);
                Console.WriteLine();
                Console.WriteLine("Compact files:");
                Console.WriteLine("{0} -p (SER|I2C|SPI) -if infile [-of outfile] [-cs \\t] [-csh] [-ds \" \"] [-ld] [-sh] [-ca] [-st]", exename);
                Console.WriteLine();
                Console.WriteLine("Best results are observed when radix in Logic is set to Hex, ASCII should also work ok-ish, others are not tested");

                return;
            }

            if (saleae_export != null)
            {
                if (saleae_dir == null)
                {
                    if (saleae_ip != "127.0.0.1" && saleae_ip.ToUpper() != "LOCALHOST")
                    {
                        Console.WriteLine("If Logic is used on a different host, an output directory available on that machine must be specified.");
                        return;
                    }

                    saleae_dir = Directory.GetCurrentDirectory();
                }
                else if(saleae_dir[1] != ':')
                {
                    // logic needs absolute directories
                    // this is very ugly but as the directory might not be discoverable (saleae may run on another machine) it's necessary
                    saleae_dir = Directory.GetCurrentDirectory() + "\\" + saleae_dir;
                }

                SaleaeClient client = null;
                try
                {
                    client = new SaleaeClient(saleae_ip, saleae_port);
                }
                catch(Exception ex)
                {
                    Console.WriteLine("Scripting server could not be reached. Is it enabled and reachable?");
                    Console.WriteLine("Guru meditation: {0}", ex.Message);
                }

                List<Analyzer> analyzers = client.GetAnalyzers();

                string[] indices = saleae_export.Split(',');

                foreach(Analyzer item in analyzers)
                {
                    if (indices.Contains(item.Index.ToString()) || indices.Contains("*"))
                    {
                        string path = string.Format("{0}\\{1}_{2}.txt", saleae_dir, item.Index, item.AnalyzerType);
                        path = path.Replace("\\\\", "\\");

                        if (!client.IsAnalyzerProcessingComplete(item.Index))
                        {
                            Console.WriteLine("Exporting Analyzer {0} ({1}) is still processing", item.Index, item.AnalyzerType);
                        }
                        else
                        {
                            Console.WriteLine("Exporting Analyzer {0} ({1}) to '{2}'", item.Index, item.AnalyzerType, path);
                            client.ExportAnalyzers(item.Index, path, false);
                        }
                    }
                }
                Console.WriteLine("done.");
                return;
            }
            else if(saleae_exportlist)
            {
                SaleaeClient client = null;
                try
                {
                    client = new SaleaeClient(saleae_ip, saleae_port);
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Scripting server could not be reached. Is it enabled and reachable?");
                    Console.WriteLine("Guru meditation: {0}", ex.Message);
                }

                List<Analyzer> analyzers = client.GetAnalyzers();
                Console.WriteLine("Available analyzers:");
                foreach (Analyzer item in analyzers)
                {
                    Console.WriteLine("{0,3}: {1}", item.Index, item.AnalyzerType);
                }
                return;
            } 
            else if(combine != null)
            {
                if(extra.Count < 2)
                {
                    Console.WriteLine("Error: Must provide at least 2 extra input files to work");
                    Console.WriteLine("Try 'serialplayer --help' for more information");
                }

                StreamWriter sw = new StreamWriter(combine, false);
                if(out_colsep_hint)
                {
                    sw.WriteLine("sep=,");
                }

                CombineFiles(extra, sw, true);
                sw.Close();
                Console.WriteLine("done.");
                return;
            }
            else
            {
                if(ifname == null)
                {
                    Console.WriteLine("No input file provided.");
                    return;
                }
                StreamReader sr = null;
                StreamWriter sw = null;
                try
                {
                    sr = new StreamReader(ifname);
                } 
                catch(Exception ex)
                {
                    Console.WriteLine("Could not open '{0}': {1}", ifname, ex.Message);
                    return;
                }

                if (ofname == null)
                {
                    // generate output file name if not given
                    ofname = Regex.Replace(ifname, @"^(.*?)(\.\w+)?$", @"\1_compact\2");
                    Console.WriteLine("No output file provided, using '{0}'", ofname);
                }
                try
                {
                    sw = new StreamWriter(ofname, false);
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Could not open '{0}': {1}", ifname, ex.Message);
                    return;
                }

                if (out_colsep == null)
                    out_colsep = ",";
                else if (out_colsep == "\\t")
                    out_colsep = "\t";

                if (out_colsep_hint)
                    sw.WriteLine("sep={0}", out_colsep);

                if (out_datasep == null) 
                    out_datasep = "";
                else if (out_datasep == "\\t")
                    out_datasep = "\t";

                if(out_colsep == out_datasep)
                {
                    Console.WriteLine("Caution: column and data separator are set to the same string, expect flawed results.");
                }

                protocol = protocol.ToUpper();

                try
                {
                    if (protocol == "I2C")
                    {
                        CompressI2C(sr, sw, out_colsep, out_datasep, !out_localdecimal, out_striphexprefix, out_convertAscii);
                    }
                    else if (protocol.StartsWith("SER"))
                    {
                        if (septime < 0)
                        {
                            Console.WriteLine("Provided separation time is only valid for DELOREAN protocol, which is not supported in this version.");
                            return;
                        }
                        else if (septime == 0)
                        {
                            Console.WriteLine("Separation time is set to 0 ms. So what is my purpose?");
                            Console.WriteLine("...but your wish is my command.");
                        }
                        CompressSerial(sr, sw, septime / 1000, out_colsep, out_datasep, !out_localdecimal, out_striphexprefix, out_convertAscii);
                    }
                    else if (protocol == "SPI")
                    {
                        CompressSpi(sr, sw, out_colsep, out_datasep, !out_localdecimal, out_striphexprefix, out_convertAscii);
                    }
                    else
                    {
                        Console.WriteLine("Protocol not recognized.");
                        return;
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("An error occured: {0}", ex.Message);
                }

                
                sr.Close();
                sw.Close();
            }
        }

        static void CsvEscape(StringBuilder stringBuilder, string columnSeparator)
        {
            string str = stringBuilder.ToString();
            if (str.Contains("\"") || str.Contains(columnSeparator))
            {
                stringBuilder = stringBuilder.Replace("\"", "\"\"");
                stringBuilder.Insert(0, "\"");
                stringBuilder.Append("\"");
            }
        }

        static void CompressI2C(StreamReader infile, StreamWriter outfile,
            string columnSeparator = "\t", string wordSeparator = " ", bool invariantDecimal = true,
            bool stripHexPrefix = false, bool convertAscii = false)
        {
            int lineno = 0;
            int lastpacketId = -1;
            double startTime = -1;
            double lastTime = 0;
            //List<string> data = new List<string>();
            StringBuilder data = new StringBuilder();
            StringBuilder acknak = new StringBuilder();
            string address = "";
            string readWrite = "";

            CultureInfo culture = invariantDecimal ? CultureInfo.InvariantCulture : CultureInfo.CurrentCulture;

            outfile.WriteLine("Start time [s]{0}End time [s]{0}Address{0}Read/Write{0}Data{0}ACK/NAK", columnSeparator);

            Action writeData = delegate
            {
                if (lastpacketId != -1)
                {
                    if (data.Length > wordSeparator.Length)
                        data.Length -= wordSeparator.Length;

                    CsvEscape(data, columnSeparator);

                    outfile.WriteLine(string.Format(culture, "{1:F13}{0}{2:F13}{0}{3}{0}{4}{0}{5}{0}{6}", columnSeparator, startTime, lastTime, address, readWrite, data, acknak));
                }
            };

            while (!infile.EndOfStream)
            {
                string line = infile.ReadLine();
                lineno++;

                // ignore header
                if (line.StartsWith("Time"))
                    continue;

                string[] parts = line.Split(',');
                // parts:
                // 0: Time
                // 1: Packet ID
                // 2: Address
                // 3: Data
                // 4: Read/Write
                // 5: ACK/NAK

                try
                {
                    double time = double.Parse(parts[0], NumberStyles.Float, CultureInfo.InvariantCulture);
                    if (startTime == -1)
                    {
                        startTime = time;
                    }
                    int packetId = -1;
                    int.TryParse(parts[1], out packetId);


                    if (lastpacketId != packetId)
                    {
                        writeData();

                        startTime = time;
                        data.Clear();
                        acknak.Clear();
                    }

                    address = parts[2];
                    if (stripHexPrefix && address.Length == 4 && address.StartsWith("0x"))
                        address = address.Substring(2);
                    if (convertAscii)
                    {
                        address = SaleaeConvertAscii(address);
                        if (address == columnSeparator)
                            address = "\"" + address + "\"";
                    }

                    string tmp = parts[3];
                    // strip prefix 0x if told so
                    if (stripHexPrefix && tmp.Length == 4 && tmp.StartsWith("0x"))
                        tmp = tmp.Substring(2);
                    // unescape ASCII if told so
                    if (convertAscii)
                        tmp = SaleaeConvertAscii(tmp);

                    data.Append(tmp);
                    data.Append(wordSeparator);

                    readWrite = parts[4];

                    switch (parts[5])
                    {
                        case "ACK": acknak.Append("A"); break;
                        case "NAK": acknak.Append("N"); break;
                        default: acknak.Append("?"); break;
                    }

                    lastTime = time;
                    lastpacketId = packetId;
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Parsing error in line {0}: {1}", lineno, ex.Message);
                }
            }

            writeData();
        }

        static double SerialWordDuration(int baudrate, int baudRate, Parity parity, int dataBits, StopBits stopBits)
        {
            double bits = ((parity == Parity.None) ? 0 : 1) + dataBits;
            switch (stopBits)
            {
                case StopBits.None: break;
                case StopBits.One: bits += 1; break;
                case StopBits.Two: bits += 2; break;
                case StopBits.OnePointFive: bits += 1.5; break;
                default:
                    break;
            }

            if (baudRate == 0) return double.PositiveInfinity;
            return bits / baudrate;
        }

        static void CompressSerial(StreamReader infile, StreamWriter outfile, double maxTimeDiff = 0,
            string columnSeparator = "\t", string wordSeparator = " ", bool invariantDecimal = true,
            bool stripHexPrefix = false, bool convertAscii = false)
        {
            int lineno = 0;
            double startTime = 0;
            double lastTime = 0;
            bool firstData = true;

            StringBuilder values = new StringBuilder();
            StringBuilder errors = new StringBuilder();

            CultureInfo culture = invariantDecimal ? CultureInfo.InvariantCulture : CultureInfo.CurrentCulture;

            outfile.WriteLine("Start time [s]{0}End time [s]{0}Values{0}Errors", columnSeparator);

            Action writeData = delegate
            {
                if (values.Length > wordSeparator.Length)
                    values.Length -= wordSeparator.Length;

                CsvEscape(values, columnSeparator);

                outfile.WriteLine(string.Format(culture, "{1:F13}{0}{2:F13}{0}{3}{0}{4}", columnSeparator, startTime, lastTime, values, errors));
            };

            while (!infile.EndOfStream)
            {
                string line = infile.ReadLine();
                lineno++;

                // ignore header
                if (line.StartsWith("Time"))
                    continue;

                string[] parts = line.Split(',');
                // parts:
                // 0: Time
                // 1: Value
                // 2: Parity Error
                // 3: Framing Error
                double time = double.Parse(parts[0], NumberStyles.Float, CultureInfo.InvariantCulture);

                if (firstData)
                {
                    firstData = false;
                    startTime = time;
                    lastTime = time;
                }

                if (time - lastTime > maxTimeDiff)
                {
                    writeData();

                    values.Clear();
                    errors.Clear();
                    startTime = time;
                }

                string tmp = parts[1]; // value

                // strip prefix 0x if told so
                if (stripHexPrefix && tmp.Length == 4 && tmp.StartsWith("0x"))
                    tmp = tmp.Substring(2);
                // unescape ASCII if told so
                if (convertAscii)
                    tmp = SaleaeConvertAscii(tmp);

                values.Append(tmp);
                values.Append(wordSeparator);

                // it's dirty but I like it
                errors.Append("-PFB"[(parts[2].Length != 0 ? 1 : 0) | (parts[3].Length != 0 ? 2 : 0)]);


                lastTime = time;
            }

            writeData();
        }

        static void CompressSpi(StreamReader infile, StreamWriter outfile,
            string columnSeparator = "\t", string wordSeparator = " ", bool invariantDecimal = true,
            bool stripHexPrefix = false, bool convertAscii = false)
        {
            int lineno = 0;
            int lastpacketId = -1;
            double startTime = -1;
            double lastTime = 0;

            StringBuilder mosiData = new StringBuilder();
            StringBuilder misoData = new StringBuilder();

            CultureInfo culture = invariantDecimal ? CultureInfo.InvariantCulture : CultureInfo.CurrentCulture;

            outfile.WriteLine("Start time [s]{0}End time [s]{0}MOSI{0}MISO", columnSeparator);

            Action writeData = delegate
            {
                if (lastpacketId != -1)
                {
                    if (mosiData.Length > wordSeparator.Length)
                    {
                        mosiData.Length -= wordSeparator.Length;
                        misoData.Length -= wordSeparator.Length;
                    }

                    CsvEscape(misoData, columnSeparator);
                    CsvEscape(mosiData, columnSeparator);


                    outfile.WriteLine(string.Format(culture, "{1:F13}{0}{2:F13}{0}{3}{0}{4}", columnSeparator, startTime, lastTime, mosiData, misoData));

                }
            };

            while (!infile.EndOfStream)
            {
                string line = infile.ReadLine();
                lineno++;

                // ignore header
                if (line.StartsWith("Time"))
                    continue;

                string[] parts = line.Split(',');
                // parts:
                // 0: Time
                // 1: Packet ID
                // 2: MOSI
                // 3: MISO

                try
                {
                    double time = double.Parse(parts[0], NumberStyles.Float, CultureInfo.InvariantCulture);
                    if (startTime == -1)
                    {
                        startTime = time;
                    }
                    int packetId = -1;
                    int.TryParse(parts[1], out packetId);


                    if (lastpacketId != packetId)
                    {
                        writeData();

                        startTime = time;
                        misoData.Clear();
                        mosiData.Clear();
                    }
                    string tmp;

                    tmp = parts[2];
                    // strip prefix 0x if told so
                    if (stripHexPrefix && tmp.Length == 4 && tmp.StartsWith("0x"))
                        tmp = tmp.Substring(2);
                    // unescape ASCII if told so
                    if (convertAscii)
                        tmp = SaleaeConvertAscii(tmp);

                    mosiData.Append(tmp);
                    mosiData.Append(wordSeparator);


                    tmp = parts[3];
                    // strip prefix 0x if told so
                    if (stripHexPrefix && tmp.Length == 4 && tmp.StartsWith("0x"))
                        tmp = tmp.Substring(2);
                    // unescape ASCII if told so
                    if (convertAscii)
                        tmp = SaleaeConvertAscii(tmp);

                    misoData.Append(tmp);
                    misoData.Append(wordSeparator);

                    lastTime = time;
                    lastpacketId = packetId;
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Parsing error in line {0}: {1}", lineno, ex.Message);
                }
            }

            writeData();
        }

        static string SaleaeConvertAscii(string str)
        {
            if (str == "COMMA")
                return ",";
            else if (str == "\\")
                return "\\\\";
            else if (str.StartsWith("'"))
            {
                if (str == "' '")
                    return " ";
                else if (str == "'")
                    return "'";
                else if (str.EndsWith("'"))
                {
                    // control characters 0 .. 8, 11 .. 12, 14 .. 32, 127 .. 255
                    int c = int.Parse(str.Substring(1, str.Length - 2));
                    switch (c)
                    {
                        case 0x00: return "\\0";
                        case 0x07: return "\\a";
                        case 0x08: return "\\b";
                        case 0x0B: return "\\v";
                        case 0x0C: return "\\f";
                        case 0x1B: return "\\e";
                        default:
                            return "\\x" + c.ToString("X2");
                    }
                }

            }
            return str;
        }


        private class CombineFileHelper
        {
            StreamReader sr = null;
            public string headers = null;
            private string colSep;
            public int colCount = 0;
            public string data = null;
            public double time = 0;
            public bool eof = false;

            public CombineFileHelper(StreamReader streamReader, string colSep)
            {
                sr = streamReader;
                this.colSep = colSep;
                headers = sr.ReadLine();
                int index = headers.IndexOf(colSep);
                //remove timestamp
                if(index != -1)
                {
                    headers = headers.Substring(index + 1);
                    colCount = headers.Length - headers.Replace(colSep, "").Length;
                }
                else
                {
                    throw new Exception("Header could not be detected");
                }
                ReadNextLine();
            }

            public bool ReadNextLine()
            {
                while (true)
                {
                    if (sr.EndOfStream)
                    {
                        eof = true;
                        data = "";
                        time = -1;
                        return false;
                    }
                    string tmp = sr.ReadLine();
                    int index = tmp.IndexOf(colSep);
                    if (index == -1)
                    {
                        // continue with next line
                        continue;
                    }
                    else
                    {
                        string timestr = tmp.Substring(0, index);
                        time = double.Parse(timestr, NumberStyles.Float, CultureInfo.InvariantCulture);
                        data = tmp.Substring(index + 1);
                        return true;
                    }
                }
            }

            public static double GetNextEntryTime(List<CombineFileHelper> files)
            {
                double mintime = double.MaxValue;
                foreach (CombineFileHelper item in files)
                {
                    if (item.eof == false && item.time < mintime)
                    {
                        mintime = item.time;
                    }
                }

                return mintime;
            }
        }

        static void CombineFiles(List<string>infiles, StreamWriter outfile, bool invariantDecimal = true)
        {
            CultureInfo culture = invariantDecimal ? CultureInfo.InvariantCulture : CultureInfo.CurrentCulture;
            string colSep = ",";

            List<CombineFileHelper> files = new List<CombineFileHelper>();
            StringBuilder header1 = new StringBuilder("File:" + colSep);
            StringBuilder header2 = new StringBuilder("Time [s]" + colSep);

            foreach (string filename in infiles)
            {
                FileInfo fi = new FileInfo(filename);
                CombineFileHelper cfh = new CombineFileHelper(new StreamReader(filename), colSep);
                files.Add(cfh);
                header1.Append(fi.Name);
                header1.Append(new string('#', cfh.colCount+1).Replace("#", colSep));
                header2.Append(cfh.headers);
                header2.Append(colSep);
            }

            outfile.WriteLine(header1);
            if(header2.Length > colSep.Length)
                header2.Length -= colSep.Length;
            outfile.WriteLine(header2);

            bool datapending = true;
            while (datapending)
            {
                datapending = false;
                double time = CombineFileHelper.GetNextEntryTime(files);
                outfile.Write(string.Format(culture, "{0:F13}{1}", time, colSep));
                for (int i = 0; i < files.Count; i++)
                {
                    
                    CombineFileHelper file = files[i];
                    if (file.time == time)
                    {
                        outfile.Write(file.data);
                        outfile.Write(colSep);
                        file.ReadNextLine();
                    }
                    else
                    {
                        outfile.Write(new string('#', file.colCount+1).Replace("#", colSep));
                    }
                    datapending |= !file.eof;
                }
                outfile.WriteLine();
            }
        }

        static void CheckSerialSeparationTime(StreamReader infile)
        {
            int lineno = 0;
            double lastTime = -1;
            int cnt = 0;
            SortedDictionary<double, int> histogram = new SortedDictionary<double, int>();

            while (!infile.EndOfStream)
            {
                string line = infile.ReadLine();
                lineno++;

                // ignore header
                if (line.StartsWith("Time"))
                    continue;

                string[] parts = line.Split(',');
                // parts:
                // 0: Time
                // 1: Value
                // 2: Parity Error
                // 3: Framing Error
                double time = double.Parse(parts[0], NumberStyles.Float, CultureInfo.InvariantCulture);
                if(lastTime < 0)
                {
                    lastTime = time;
                    continue;
                }

                double tdiff = time - lastTime;
                cnt++;

                if(histogram.ContainsKey(tdiff))
                {
                    histogram[tdiff]++;
                }
                else
                {
                    histogram.Add(tdiff, 1);
                }
                lastTime = time;
            }

            foreach(KeyValuePair<double, int> pair in histogram)
            {
                Console.WriteLine("{0:F9} ms: {1} ({2:F2} %)", pair.Key * 1000, pair.Value, ((double)pair.Value / cnt) * 100);
            }
        }
    }
}
